@@ -263,6 +263,7 @@ bgp_prepare_capabilities(struct bgp_conn *conn)
   conn->local_caps = caps;
 
   caps->as4_support = p->cf->enable_as4;
+  caps->as4_number = p->public_as;
   caps->ext_messages = p->cf->enable_extended_messages;
   caps->route_refresh = p->cf->enable_refresh;
   caps->enhanced_refresh = p->cf->enable_refresh && p->cf->enable_enhanced_refresh;
@@ -325,10 +326,8 @@ bgp_prepare_capabilities(struct bgp_conn *conn)
 }
 
 static byte *
-bgp_write_capabilities(struct bgp_conn *conn, byte *buf)
+bgp_write_capabilities(struct bgp_caps *caps, byte *buf)
 {
-  struct bgp_proto *p = conn->bgp;
-  struct bgp_caps *caps = conn->local_caps;
   struct bgp_af_caps *ac;
   byte *buf_head = buf;
   byte *data;
@@ -411,7 +410,7 @@ bgp_write_capabilities(struct bgp_conn *conn, byte *buf)
   {
     *buf++ = 65;		/* Capability 65: Support for 4-octet AS number */
     *buf++ = 4;			/* Capability data length */
-    put_u32(buf, p->public_as);
+    put_u32(buf, caps->as4_number);
     buf += 4;
   }
 
@@ -853,7 +852,7 @@ bgp_create_open(struct bgp_conn *conn, byte *buf)
   {
     /* Prepare local_caps and write capabilities to buffer */
     byte *pos = buf+12;
-    byte *end = bgp_write_capabilities(conn, pos);
+    byte *end = bgp_write_capabilities(conn->local_caps, pos);
     uint len = end - pos;
 
     if (len < 254)
@@ -1085,12 +1084,16 @@ bgp_apply_next_hop(struct bgp_parse_state *s, rta *a, ip_addr gw, ip_addr ll)
   if (c->cf->gw_mode == GW_DIRECT)
   {
     neighbor *nbr = NULL;
+    uint nb_flags = p->cf->onlink ? NEF_ONLINK : 0;
+    struct iface *default_iface = p->cf->onlink ? p->neigh->iface : NULL;
 
     /* GW_DIRECT -> single_hop -> p->neigh != NULL */
-    if (ipa_nonzero2(gw))
-      nbr = neigh_find(&p->p, gw, NULL, 0);
+    if ((c->cf->next_hop_prefer == NHP_GLOBAL) && ipa_nonzero2(gw))
+      nbr = neigh_find(&p->p, gw, default_iface, nb_flags);
     else if (ipa_nonzero(ll))
-      nbr = neigh_find(&p->p, ll, p->neigh->iface, 0);
+      nbr = neigh_find(&p->p, ll, p->neigh->iface, nb_flags);
+    else if (ipa_nonzero2(gw))
+      nbr = neigh_find(&p->p, gw, default_iface, nb_flags);
     else
       WITHDRAW(BAD_NEXT_HOP " - zero address");
 
@@ -1103,6 +1106,7 @@ bgp_apply_next_hop(struct bgp_parse_state *s, rta *a, ip_addr gw, ip_addr ll)
     a->dest = RTD_UNICAST;
     a->nh.gw = nbr->addr;
     a->nh.iface = nbr->iface;
+    a->nh.flags = nbr->flags & NEF_ONLINK ? RNF_ONLINK : 0;
     a->igp_metric = c->cf->cost;
   }
   else /* GW_RECURSIVE */
